package cz.jzitnik.game.core.autotransient;

import cz.jzitnik.game.annotations.AutoTransient;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serial;
import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;

public abstract class AutoTransientSupport implements Serializable {

    @Serial
    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        reinitializeAutoTransients();
    }

    protected void reinitializeAutoTransients() {
        for (Field field : this.getClass().getDeclaredFields()) {
            if (field.isAnnotationPresent(AutoTransient.class)) {
                field.setAccessible(true);
                AutoTransient annotation = field.getAnnotation(AutoTransient.class);
                try {
                    Object value;

                    // Use initializer if provided
                    Class<? extends AutoTransientInitializer<?>> initializerClass = annotation.initializer();
                    if (!initializerClass.equals(DefaultInitializer.class)) {
                        AutoTransientInitializer<?> initializer = initializerClass.getDeclaredConstructor().newInstance();
                        value = initializer.initialize(this);
                    } else {
                        // Fallback to default instantiation
                        value = instantiateField(field.getType());
                    }

                    if (value != null) {
                        field.set(this, value);
                    }

                } catch (Exception e) {
                    throw new RuntimeException("Failed to reinitialize @AutoTransient field: " + field.getName(), e);
                }
            }
        }
    }

    protected Object instantiateField(Class<?> clazz) {
        try {
            // Try constructor with (this) reference first
            Constructor<?> constructor = clazz.getDeclaredConstructor(this.getClass());
            return constructor.newInstance(this);
        } catch (NoSuchMethodException e) {
            // Fallback to default constructor
            try {
                return clazz.getDeclaredConstructor().newInstance();
            } catch (Exception ex) {
                ex.printStackTrace();
                return null;
            }
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }
}
